"""
@file
@brief Shorcuts to plotting functions.
"""

from .plotting_benchmark import plot_benchmark_metrics  # pylint: disable=W0611
from .plotting_validate_graph import plot_validate_benchmark  # pylint: disable=W0611
from .plotting_onnx import plot_onnx  # pylint: disable=W0611
